@@ -887,7 +887,7 @@ int meta_ascmsgtoav(META *m, char *buf, ssize_t len,
 
 	/* Find the specified item in the dictionary */
 	msg(F_PROC, L_DEBUG, "- got attribute '%.*s'\n", rl, atr);
-	av->i = meta_getitembyspec(m, atr, rl);
+	av->i = meta_getitembyextspec(m, atr, rl, &av->ref);
 	if (!av->i) { msg(F_PROC, L_ERR, "meta_ascmsgtoav: ERROR: Unknown "
 					 "attribute '%.*s'!\n", rl, atr);
 		      free(av); return -1; }
@@ -928,13 +928,11 @@ int meta_ascmsgtoav(META *m, char *buf, ssize_t len,
 		return 0;
 	}
 
-	/* No hex values; do not allow empty values */
-	if (c >= end) goto meta_ascmsgtoav_invpair;
-
 	/* Parse value according to item's data type */
 	switch(av->i->val_type) {
 	  case MT_INTEGER:
 	  case MT_DATE:
+		if (c >= end) goto meta_ascmsgtoav_invpair;
 		av->l = meta_atoord(c, end - c, 0, 0, &rl, 10);
 		if (rl) break;
 
@@ -945,11 +943,13 @@ int meta_ascmsgtoav(META *m, char *buf, ssize_t len,
 		break;
 
 	  case MT_IPADDR:
+		if (c >= end) goto meta_ascmsgtoav_invpair;
 		av->l = meta_atoip(c, end - c, 0, 0, &rl);
 		if (!rl) goto meta_ascmsgtoav_invpair;
 		break;
 
 	  default:
+		if (c >= end) break;	    /* empty strings */
 	  	/* Note: we allocate as many bytes as there are left in the
 		   input message. This is safe, because prttoa can never 
 		   output more data than we feed it, only less. Guaranteed. */
@@ -1066,9 +1066,9 @@ int meta_binmsgtoav(META *m, U_INT32_T *buf, ssize_t len,
 
 ssize_t meta_avtomsg(META *m, META_AV *avlst, 
 		     char *buf, ssize_t buflen, 
-		     int flags, META_AV *acl, U_INT32_T magic)
+		     int flags, U_INT32_T binmagic)
 {
-	META_AV *av, *aclav;
+	META_AV *av;
 	META_ORD l, tmpord;
 	META_VND *vnd;
 	META_VAL *v;
@@ -1084,14 +1084,6 @@ ssize_t meta_avtomsg(META *m, META_AV *avlst,
 			if ((av->flags & AV_DEL) && 
 			    (flags & AVMSG_ONESHOT) == 0) continue;
 
-			/* Check if av occurs in ACL, if any */
-			if (acl) {
-				for(aclav = acl; 
-				    aclav && aclav->i != av->i;
-				    aclav = aclav->next);
-				if (!aclav) continue;
-			}
-
 			/* Determine output length first (8 if anonymous ord) */
 			if (av->p) l = av->l;
 			else l = av->i ? (av->i->val_size > 0 ? av->i->val_size
@@ -1138,7 +1130,7 @@ ssize_t meta_avtomsg(META *m, META_AV *avlst,
 		l = o - buf;
 
 		/* Put in request/reply header magic as given by caller */
-		putord(buf, 4, magic);
+		putord(buf, 4, binmagic);
 		putord(buf + 4, 4, l);
 		return l;
 	}
@@ -1149,14 +1141,6 @@ ssize_t meta_avtomsg(META *m, META_AV *avlst,
 		if ((av->flags & AV_DEL) && 
 		    (flags & AVMSG_ONESHOT) == 0) continue;
 
-		/* Check if av occurs in ACL, if any */
-		if (acl) {
-			for(aclav = acl; 
-			    aclav && aclav->i != av->i;
-			    aclav = aclav->next);
-			if (!aclav) continue;
-		}
-
 		/* Test if we're adding tabs in front */
 		if (flags & AVMSG_ADDTAB) {
 			buflen--; if (buflen < 0) break; 
@@ -1366,6 +1350,23 @@ void meta_addav(META_AV **head, META_AV **tail, META_AV *rel, int before,
 }
 
 
+/* Convenient wrapper for malloc, memset and meta_addav */
+
+META_AV *meta_addnewav(META_AV **head, META_AV **tail, 
+			      META_AV *rel, int before, 
+			      META_ITEM *i, char *p, META_ORD l)
+{
+	META_AV *av;
+
+	av = (META_AV *)malloc(sizeof(META_AV)); if (!av) return 0;
+	memset(av, 0, sizeof(META_AV));
+	av->i = i; av->p = p; av->l = l; 
+
+	meta_addav(head, tail, rel, before, av);
+	return av;
+}
+
+
 /* Remove an AV item from a list, without freeing it */
 
 void meta_remav(META_AV **head, META_AV **tail, META_AV *av)
@@ -1443,8 +1444,7 @@ void meta_printav(META *m, META_AV *av, int depth)
 	}
 
 	n = meta_avtomsg(m, av, o, buflen - 1, AVMSG_ONESHOT + AVMSG_ASCII + 
-			 AVMSG_ADDTAB + AVMSG_ADDSPACES + AVMSG_NAMEDCONST, 
-			 0, 0);
+			 AVMSG_ADDTAB + AVMSG_ADDSPACES + AVMSG_NAMEDCONST, 0);
 	buflen -= n; ret += n;
 
 	buf[ret] = 0; 
