@@ -37,7 +37,7 @@ char langcompile_id[] = "LANGCOMPILE - Copyright (C) 2001 Emile van Bergen.";
 #include <evblib/db/list.h>	/* for LIST_ */
 
 
-/* (Shortened) type definitions for use in the standard operator table */
+/* (Shortened) type definitions for use in the operator table */
 
 #define NON		-1	/* As context: no type
 				   	(expecting term or unary prefix op)
@@ -87,87 +87,45 @@ typedef struct stdop {
  */
 
 
-/* The standard operators we know */
+/* Standard operators tested before tokenizing - no named identifiers or
+ * operators may start with these strings, because they will be shadowed */
 
 static STDOP stdops[] = {
 
-	/* Halting 'terms' - don't recurse, apply only in 'NON' and pretend
-	 * they return an integer on the stack, so that the comma will work. */
-	{ pstr("halt"),	 	NON, 32, L,   ANY, NON,   INT, OP_HALT },
-	{ pstr("abort"),	NON, 32, L,   ANY, NON,   INT, OP_ABORT },
-	{ pstr("accept"),	NON, 32, L,   ANY, NON,   INT, OP_ACCEPT },
-	{ pstr("reject"),	NON, 32, L,   ANY, NON,   INT, OP_REJECT },
-	{ pstr("acctresp"),	NON, 32, L,   ANY, NON,   INT, OP_ACCTRESP },
-
-	/* Unary postfix operators */
-	{ pstr("exists"),	ANY, 30, L,   ANY, NON,   INT, OP_EXISTS },
-	/* Typecasts / auto-conversion (uses autodetected format for ints) */
-	{ pstr("toint"),	ANY, 30, L,   INT, NON,   INT, OP_NOP },
-	{ pstr("toip"),	 	ANY, 30, L,   IPA, NON,   IPA, OP_NOP },
-	{ pstr("todate"),	ANY, 30, L,   DAT, NON,   DAT, OP_NOP },
-	{ pstr("tostr"),	ANY, 30, L,   STR, NON,   STR, OP_NOP },
-	/* Convert specific string format to int (oct/dec/hex/raw -> int) */
-	{ pstr("fromoct"),	ANY, 30, L,   STR, NON,   INT, OP_OCTSTR2ORD },
-	{ pstr("fromdec"),	ANY, 30, L,   STR, NON,   INT, OP_DECSTR2ORD },
-	{ pstr("fromhex"),	ANY, 30, L,   STR, NON,   INT, OP_HEXSTR2ORD },
-	{ pstr("fromraw"),	ANY, 30, L,   STR, NON,   INT, OP_RAWSTR2ORD },
-	/* Convert int to specific string format (int -> oct/dec/hex/raw) */
-	{ pstr("tooct"),	ANY, 30, L,   INT, NON,   STR, OP_ORD2OCTSTR },
-	{ pstr("todec"),	ANY, 30, L,   INT, NON,   STR, OP_ORD2DECSTR },
-	{ pstr("tohex"),	ANY, 30, L,   INT, NON,   STR, OP_ORD2HEXSTR },
-	{ pstr("toraw"),	ANY, 30, L,   INT, NON,   STR, OP_ORD2RAWSTR },
-
 	/* Unary prefix operators. All have same precedence and associate 
 	 * right, otherwise behaviour would be quite counterintuitive; the
-	 * one closest to the term is always evaluated first. */
-	{ pstr("delall"),	NON, 28, R,   ANY, ANY,   INT, OP_DELALLAV },
-	{ pstr("del"),	 	NON, 28, R,   ANY, ANY,   INT, OP_DELAV },
-	{ pstr("moveall"),	NON, 28, R,   ANY, ANY,   INT, OP_MOVEALLAV },
-	{ pstr("all"),	 	NON, 28, R,   ANY, ANY,   STR, OP_JOINSTR },
-	{ pstr("return"), 	NON, 28, R,   ANY, INT,   INT, OP_RETFUNC },
+	 * one closest to the term should always be evaluated first. */
+
 	{ pstr("~"),		NON, 28, R,   ANY, INT,   INT, OP_NOT },
 	{ pstr("-"),		NON, 28, R,   ANY, INT,   INT, OP_NEG },
-	{ pstr("dictencode"),	NON, 28, R,   ANY, ANY,   STR, OP_DICTENC },
 	{ pstr("!"),		NON, 28, R,   ANY, ANY,   INT, OP_BOOLNOT },
-	{ pstr("not"),	 	NON, 28, R,   ANY, ANY,   INT, OP_BOOLNOT },
-	{ pstr("no"),		NON, 28, R,   ANY, ANY,   INT, OP_BOOLNOT },
-	{ pstr("random"),	NON, 28, R,   ANY, ANY,   STR, OP_RANDOM },
-	{ pstr("md5"),	 	NON, 28, R,   ANY, STR,   STR, OP_MD5 },
-	{ pstr("hex"),	 	NON, 28, R,   ANY, STR,   STR, OP_HEX },
-	{ pstr("upper"),	NON, 28, R,   ANY, STR,   STR, OP_UPPER },
-	{ pstr("lower"),	NON, 28, R,   ANY, STR,   STR, OP_LOWER },
+
 	/* Binary operators */
+
 	{ pstr("*"),		ANY, 25, L,   INT, INT,   INT, OP_MUL },
 	{ pstr("/"),		IPA, 25, L,   ANY, INT,   IPA, OP_CIDRMASK },
 	{ pstr("/"),		ANY, 25, L,   INT, INT,   INT, OP_DIV },
 	{ pstr("%"),		ANY, 25, L,   INT, INT,   INT, OP_MOD },
+
 	{ pstr("+"),		DAT, 24, L,   ANY, INT,   DAT, OP_ADD },
 	{ pstr("+"),		IPA, 24, L,   ANY, INT,   IPA, OP_ADD },
 	{ pstr("+"),		ANY, 24, L,   INT, INT,   INT, OP_ADD },
 	{ pstr("-"),		ANY, 24, L,   INT, INT,   INT, OP_SUB },
+
 	{ pstr(">>"),		ANY, 23, L,   INT, INT,   INT, OP_SHR },
 	{ pstr("<<"),		ANY, 23, L,   INT, INT,   INT, OP_SHL },
+
 	{ pstr("^"),		INT, 22, L,   ANY, INT,   INT, OP_XOR },
 	{ pstr("^"),		IPA, 22, L,   ANY, IPA,   IPA, OP_XOR },
 	{ pstr("^"),		DAT, 22, L,   ANY, DAT,   DAT, OP_XOR },
 	{ pstr("^"),		STR, 22, L,   ANY, STR,   STR, OP_XORSTR },
+
 	{ pstr("&"),		ANY, 21, L,   INT, INT,   INT, OP_AND },
+
 	{ pstr("|"),		ANY, 20, L,   INT, INT,   INT, OP_OR },
-	{ pstr("asmac"),	ANY, 18, L,   STR, STR,   STR, OP_STR2MAC },
-	{ pstr("papdecrypt"),	ANY, 18, L,   STR, STR,   STR, OP_PAPDECR },
-	{ pstr("papencrypt"),	ANY, 18, L,   STR, STR,   STR, OP_PAPENCR },
-	{ pstr("unixcrypt"),	ANY, 18, L,   STR, STR,   STR, OP_UNIXCRYPT },
-	{ pstr("hmacmd5"),	ANY, 18, L,   STR, STR,   STR, OP_HMACMD5 },
-	{ pstr("dictdecode"),	ANY, 18, L,   STR, ANY,   INT, OP_DICTDEC },
-	/* Convert to run-time specified format (only implemented for dates) */
-	{ pstr("as"),		ANY, 18, L,   DAT, STR,   STR, OP_ORD2DFMSTR },
-	{ pstr("beforefirst"),	ANY, 16, L,   STR, STR,   STR, OP_BF },
-	{ pstr("afterfirst"),	ANY, 16, L,   STR, STR,   STR, OP_AF },
-	{ pstr("beforelast"),	ANY, 16, L,   STR, STR,   STR, OP_BL },
-	{ pstr("afterlast"),	ANY, 16, L,   STR, STR,   STR, OP_AL },
-	{ pstr("firstof"),	ANY, 16, R,   INT, STR,   STR, OP_FO },
-	{ pstr("lastof"),	ANY, 16, R,   INT, STR,   STR, OP_LO },
+
 	{ pstr("."),		ANY, 14, L,   STR, STR,   STR, OP_CONCAT },
+
 	{ pstr(">="),		STR, 12, L,   ANY, STR,   INT, OP_GESTR },
 	{ pstr(">="),		ANY, 12, L,   INT, INT,   INT, OP_GE },
 	{ pstr("<="),		STR, 12, L,   ANY, STR,   INT, OP_LESTR },
@@ -176,10 +134,12 @@ static STDOP stdops[] = {
 	{ pstr(">"),		ANY, 12, L,   INT, INT,   INT, OP_GT },
 	{ pstr("<"),		STR, 12, L,   ANY, STR,   INT, OP_LTSTR },
 	{ pstr("<"),		ANY, 12, L,   INT, INT,   INT, OP_LT },
+
 	{ pstr("!="),		STR, 11, L,   ANY, STR,   INT, OP_NESTR },
 	{ pstr("!="),		ANY, 11, L,   ANY, ANY,   INT, OP_NE },
 	{ pstr("=="),		STR, 11, L,   ANY, STR,   INT, OP_EQSTR },
 	{ pstr("=="),		ANY, 11, L,   ANY, ANY,   INT, OP_EQ },
+
 	{ pstr("="),		INT,  7, R,   ANY, INT,   INT, OP_ADDAV },
 	{ pstr("="),		IPA,  7, R,   ANY, IPA,   IPA, OP_ADDAV },
 	{ pstr("="),		DAT,  7, R,   ANY, DAT,   DAT, OP_ADDAV },
@@ -188,14 +148,84 @@ static STDOP stdops[] = {
 	{ pstr(":="),		IPA,  7, R,   ANY, IPA,   IPA, OP_REPLACEAV },
 	{ pstr(":="),		DAT,  7, R,   ANY, DAT,   DAT, OP_REPLACEAV },
 	{ pstr(":="),		STR,  7, R,   ANY, STR,   STR, OP_REPLACEAV },
-	{ pstr("pokedwith"),	STR,  7, R,   ANY, STR,   STR, OP_POKEAV },
-	{ pstr(","),		NON,  1, L,   ANY, NON,   NON, OP_NOP },
-	{ pstr(","),		ANY,  1, L,   ANY, NON,   NON, OP_POP }
 };
 
 #define STDOP_CNT	(sizeof(stdops) / sizeof(STDOP))
 
 
+/* Standard operators tested after tokenizing by taking A-Za-z0-9:- span */
+
+static STDOP namedops[] = {
+
+	/* Halting 'terms' - don't recurse, apply only in 'NON' and pretend
+	 * they return an integer on the stack, so that the comma will work. */
+
+	{ pstr("halt"),	 	NON, 32, L,   ANY, NON,   INT, OP_HALT },
+	{ pstr("abort"),	NON, 32, L,   ANY, NON,   INT, OP_ABORT },
+	{ pstr("accept"),	NON, 32, L,   ANY, NON,   INT, OP_ACCEPT },
+	{ pstr("reject"),	NON, 32, L,   ANY, NON,   INT, OP_REJECT },
+	{ pstr("acctresp"),	NON, 32, L,   ANY, NON,   INT, OP_ACCTRESP },
+
+	/* Unary postfix operators */
+
+	{ pstr("exists"),	ANY, 30, L,   ANY, NON,   INT, OP_EXISTS },
+	/* typecasts / auto-conversion (uses autodetected format for ints) */
+	{ pstr("toint"),	ANY, 30, L,   INT, NON,   INT, OP_NOP },
+	{ pstr("toip"),	 	ANY, 30, L,   IPA, NON,   IPA, OP_NOP },
+	{ pstr("todate"),	ANY, 30, L,   DAT, NON,   DAT, OP_NOP },
+	{ pstr("tostr"),	ANY, 30, L,   STR, NON,   STR, OP_NOP },
+	/* convert specific string format to int (oct/dec/hex/raw -> int) */
+	{ pstr("fromoct"),	ANY, 30, L,   STR, NON,   INT, OP_OCTSTR2ORD },
+	{ pstr("fromdec"),	ANY, 30, L,   STR, NON,   INT, OP_DECSTR2ORD },
+	{ pstr("fromhex"),	ANY, 30, L,   STR, NON,   INT, OP_HEXSTR2ORD },
+	{ pstr("fromraw"),	ANY, 30, L,   STR, NON,   INT, OP_RAWSTR2ORD },
+	/* convert int to specific string format (int -> oct/dec/hex/raw) */
+	{ pstr("tooct"),	ANY, 30, L,   INT, NON,   STR, OP_ORD2OCTSTR },
+	{ pstr("todec"),	ANY, 30, L,   INT, NON,   STR, OP_ORD2DECSTR },
+	{ pstr("tohex"),	ANY, 30, L,   INT, NON,   STR, OP_ORD2HEXSTR },
+	{ pstr("toraw"),	ANY, 30, L,   INT, NON,   STR, OP_ORD2RAWSTR },
+
+	/* Unary prefix operators. All have same precedence and associate 
+	 * right, otherwise behaviour would be quite counterintuitive; the
+	 * one closest to the term should always be evaluated first. */
+
+	{ pstr("delall"),	NON, 28, R,   ANY, ANY,   INT, OP_DELALLAV },
+	{ pstr("del"),	 	NON, 28, R,   ANY, ANY,   INT, OP_DELAV },
+	{ pstr("moveall"),	NON, 28, R,   ANY, ANY,   INT, OP_MOVEALLAV },
+	{ pstr("all"),	 	NON, 28, R,   ANY, ANY,   STR, OP_JOINSTR },
+	{ pstr("return"), 	NON, 28, R,   ANY, INT,   INT, OP_RETFUNC },
+	{ pstr("dictencode"),	NON, 28, R,   ANY, ANY,   STR, OP_DICTENC },
+	{ pstr("not"),	 	NON, 28, R,   ANY, ANY,   INT, OP_BOOLNOT },
+	{ pstr("no"),		NON, 28, R,   ANY, ANY,   INT, OP_BOOLNOT },
+	{ pstr("random"),	NON, 28, R,   ANY, ANY,   STR, OP_RANDOM },
+	{ pstr("md5"),	 	NON, 28, R,   ANY, STR,   STR, OP_MD5 },
+	{ pstr("hex"),	 	NON, 28, R,   ANY, STR,   STR, OP_HEX },
+	{ pstr("upper"),	NON, 28, R,   ANY, STR,   STR, OP_UPPER },
+	{ pstr("lower"),	NON, 28, R,   ANY, STR,   STR, OP_LOWER },
+
+	/* Binary operators */
+
+	{ pstr("asmac"),	ANY, 18, L,   STR, STR,   STR, OP_STR2MAC },
+	{ pstr("papdecrypt"),	ANY, 18, L,   STR, STR,   STR, OP_PAPDECR },
+	{ pstr("papencrypt"),	ANY, 18, L,   STR, STR,   STR, OP_PAPENCR },
+	{ pstr("unixcrypt"),	ANY, 18, L,   STR, STR,   STR, OP_UNIXCRYPT },
+	{ pstr("hmacmd5"),	ANY, 18, L,   STR, STR,   STR, OP_HMACMD5 },
+	{ pstr("dictdecode"),	ANY, 18, L,   STR, ANY,   INT, OP_DICTDEC },
+	{ pstr("as"),		ANY, 18, L,   DAT, STR,   STR, OP_ORD2DFMSTR },
+
+	{ pstr("beforefirst"),	ANY, 16, L,   STR, STR,   STR, OP_BF },
+	{ pstr("afterfirst"),	ANY, 16, L,   STR, STR,   STR, OP_AF },
+	{ pstr("beforelast"),	ANY, 16, L,   STR, STR,   STR, OP_BL },
+	{ pstr("afterlast"),	ANY, 16, L,   STR, STR,   STR, OP_AL },
+	{ pstr("firstof"),	ANY, 16, R,   INT, STR,   STR, OP_FO },
+	{ pstr("lastof"),	ANY, 16, R,   INT, STR,   STR, OP_LO },
+
+	{ pstr("pokedwith"),	STR,  7, R,   ANY, STR,   STR, OP_POKEAV }
+};
+
+#define NAMEDOP_CNT	(sizeof(namedops) / sizeof(STDOP))
+
+
 /*
  * FUNCTIONS
  */
@@ -225,7 +255,7 @@ static int cvtop(int srctype, int dsttype)
 }
 
 
-/* Add a function to the function list */
+/* Add a function at the head of the function list (later overrides earlier) */
 
 static void func_add(FUNC_T **fhead, FUNC_T **ftail, char *name, ssize_t namel,
 		     INSN *code)
@@ -236,7 +266,7 @@ static void func_add(FUNC_T **fhead, FUNC_T **ftail, char *name, ssize_t namel,
 	memset(f, 0, sizeof(FUNC_T));
 	f->namel = cptstr(f->name, sizeof(f->name), name, namel);
 	f->code = code;
-	LIST_ADDTAIL(fhead, ftail, f);
+	LIST_ADDHEAD(fhead, ftail, f);
 }
 
 
@@ -275,24 +305,24 @@ static ssize_t reccompilefile(META *m, IFACE *ifaces,
 #define GETIP(rl)	meta_atoip(*source, *sourcel, 0, 0, (rl))
 #define GETORDN(b,r,n)  meta_atoord(*source, MIN(*sourcel, (n)), 0, 0, (r), (b))
 #define GETORD(ba,rl)	meta_atoord(*source, *sourcel, 0, 0, (rl), (ba))
-#define SPN()		memspn(*source, *sourcel, pstr("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789:-"))
+#define TOKEN()		memspn(*source, *sourcel, pstr("ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789:-"))
 #define FEQ(s, l)	((l) > *sourcel || memcmp(*source, (s), (l)))  /*begin*/
 #define EQ(sl, s, l)	((l) - (sl) || memcmp(*source, (s), (l)))      /*to sl*/
 #define DBGCVT(n)	dbg_cvtstr(*source, (n))
 
 
 static ssize_t reccompile(META *m, IFACE *ifaces, 
-			  char **source, ssize_t *sourcel,
+			  char **source, ssize_t *sourcel, 
 			  INSN **buf, ssize_t bufsize, 
 			  int precedence, int req_ctx, 
-			  META_ITEM *ctx_item, int *ret_ctx,
-			  char *fullpath, ssize_t pathl,
+			  META_ITEM *ctx_item, int *ret_ctx, int allow_non, 
+			  char *fullpath, ssize_t pathl, 
 			  FUNC_T **funchead, FUNC_T **functail)
 {
 	signed char *s, c;
 	INSN *o, *saved_o;
 	ssize_t l, len, ret;
-	int context, n;
+	int context, n, e;
 	META_VAL *v;
 	STDOP *so;
 	IFACE *ifc;
@@ -302,16 +332,25 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 		msg(F_MISC, L_ERR, "reccompile: BUG: Invalid argument(s)!\n");
 		return 0;
 	}
+	D1(msg(F_LANG, L_DEBUG, "reccompile: c=%c, prec=%d, req_ctx=%d, ctx_item=%s\n", C, precedence, req_ctx, ctx_item ? ctx_item->name:"-"));
+
+	/* We always start each recursive invocation in context 'NON', ie.
+	 * expecting a term from the source. And we always return a term, or
+	 * NON if that's what the parent required (req_ctx NON) or allowed
+	 * (allow_non true).
+	 *
+	 * Whether or not we process the operator after the term we find on the
+	 * current level, or whether we return first to leave it to the parent,
+	 * depends on its precedence level relative to the parent's. */
 
 	context = NON;
 	o = *buf;
 	ret = 0;
-	D1(msg(F_LANG, L_DEBUG, "reccompile: c=%c, prec=%d, req_ctx=%d, "
-				"context=%d, ctx_item=%s\n",
-	       C, precedence, req_ctx, context, ctx_item ? ctx_item->name:"-"));
 
-	/* We leave guard space of min. 16 insns so that we can safely emit
-	 * inside this loop. */
+	/* We leave guard space of min. 16 INSNs so that we can safely emit
+	 * a couple of them inside the loop, only testing for long ones like
+	 * PUSHSTRs. */
+
 	while(*sourcel > 0 && ret < bufsize - (16 * sizeof(INSN))) {
 
 		/* Get first character */
@@ -340,14 +379,15 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 		 * dictionaries contain stuff like 1200-75-BPS (yuck!) */
 
 		len = -1;
-		if (ctx_item &&
-		    (len = SPN()) &&
-		    (v = meta_getvalbyname(m, ctx_item, *source, len))) {
-			S(len);
-			context = INT;
-			o->imm.ord = v->nr;
-			o->op = OP_PUSHINT; o++; ret += sizeof(INSN);
-			continue;
+		if (ctx_item && (len = TOKEN())) {
+			if (I(len - 1) == ':') len--; 
+			if ((v = meta_getvalbyname(m, ctx_item, *source, len))){
+				S(len);
+				context = INT;
+				o->imm.ord = v->nr;
+				o->op = OP_PUSHINT; o++; ret += sizeof(INSN);
+				continue;
+			}
 		}
 
 		/* Numeric term. Also supports dotted-decimal IP addresses. */
@@ -397,94 +437,116 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 		   undisturbed until we meet the first closing one. At that
 		   point, return until we're up at the precedence level the 
 		   opening one brought us, then skip it. */
+
 		if (c == '(' && context == NON) {
 			S(1);
 			/* We don't pass the attribute context beyond ( */
-			/* Also, our returned type is determined by the
-			 * subexpression. Don't autoconvert. */
+			/* The type of this term is determined by the type
+			 * of the subexpression. Don't autoconvert; we'll
+			 * do that at this level if we needed something
+			 * else. Pass parent's allow_non to subexpression. */
+
 			l = reccompile(m, ifaces, source, sourcel, 
-				       &o, bufsize - ret, 0,
-				       ANY, 0, &context, 
+				       &o, bufsize - ret, 
+				       0, ANY, 0, &context, allow_non,
 				       fullpath, pathl, funchead, functail);
 			if (l == -1) return -1;
 			ret += l;
 			continue;
 		}
-		if (c == ')') {
+		if (c == ')' && (context != NON || allow_non)) {
 			if (precedence > 0) break;
 			S(1);
 			break;
 		}
 
-		/* These boolean operators use short-circuit evaluation.
-		   They return the last evaluated term as the result. This
-		   makes constructs possible like ArgA || ArgB, returning
-		   ArgA if that has a true value, or ArgB if ArgA is false. */
-		if (((c == '&' && I(1) == '&') ||
-		     (c == 'a' && I(2) == 'd' && I(1) == 'n')) && 
-		    context != NON) { 
-			if (precedence > 5) break;
-			S(2 + (c == 'a'));
-			saved_o = o;
-			o->op = OP_JMPZ; o++;
-			o->op = OP_POP; o++; 
-			/* Funky feature: 'and' does not require the left hand
-			 * term to be any specific type, because to the VM, a
-			 * false int looks exactly like a false str, so if we
-			 * return that, then that's what the stack has anyway.
-			 * In all cases we can set the context after an 'and'
-			 * to the right hand's type. */
-			context = req_ctx; ctx_item = 0;
-			l = reccompile(m, ifaces, source, sourcel,
-				       &o, bufsize - ret, 5,
-				       ANY, ctx_item, &context,
-				       fullpath, pathl, funchead, functail);
-			if (l == -1) return -1;	
-			saved_o->imm.d.disp = l / sizeof(INSN) + 1;
-			ret += l + 2 * sizeof(INSN);
-			continue;
-		}
+		/* A comma is a no-op (we don't even emit a NOP) if we're still
+		 * expecting a term. Otherwise, it's a POP that we'll replace
+		 * with a NOP if the right hand side didn't give us any term.
+		 * Why this complexity? It allows (3,) to return 3, making the
+		 * language much more user friendly with respect to comma 
+		 * placement. */
 
-		if (((c == '|' && I(1) == '|') ||
-		     (c == 'o' && I(1) == 'r')) && context != NON) { 
-			if (precedence > 4) break;
-			S(2);
+		if (c == ',') {
+		    if (precedence >= 1) break;
+		    S(1);
+		    ctx_item = 0;		/* Comma always resets 
+						   attribute context */
+		    if (context != NON) {
+
+			/* Emit POP, possibly removed after reccompile */
 			saved_o = o;
-			o->op = OP_JMPNZ; o++;
-			o->op = OP_POP; o++; 
-			/* Or causes rhs to be converted to lhs' type */
+			o->op = OP_POP; o++; ret += sizeof(INSN);
+
+			/* We don't pass the attribute context; we don't
+			 * autoconvert but want to know the returned context,
+			 * we allow_non. */
+
 			l = reccompile(m, ifaces, source, sourcel,
 				       &o, bufsize - ret, 
-				       4, context, ctx_item, 0,
+				       1, ANY, 0, &n, 1,
 				       fullpath, pathl, funchead, functail);
-			if (l == -1) return -1;	
-			saved_o->imm.d.disp = l / sizeof(INSN) + 1;
-			ret += l + 2 * sizeof(INSN);
-			continue;
+			if (l == -1) return -1;
+			ret += l;
+
+			/* If reccompile returned nothing at all, we can remove
+			 * the POP by backing up. If it returned some INSNs but
+			 * didn't push a term (in whatever way), we replace the
+			 * POP with a NOP (and keep the current term and its
+			 * context). In all other cases, we keep the POP and
+			 * take on the returned context. */
+
+			if (!l) o--, ret -= sizeof(INSN);
+			else if (n == NON) saved_o->op = OP_NOP;
+			else context = n;
+		    }
+		    continue;
 		}
 
 		/*
-		 * STANDARD OPERATORS
+		 * STANDARD AND NAMED OPERATORS
 		 */
 
 		/* Search in standard operator table for first match of name
 		   and context (or any context). The table must be ordered,
 		   with the more specific names and contexts first; the amount
-		   of text we test depends entirely on the operator. So be sure
-		   not to clash with the start of dictionary items or
-		   interfaces; a unary prefix operator 'in' (ie. valid in
-		   context 'NON') would clobber the dictionary item 'int', for
-		   example! We'll fix that by splitting named and symbolic
-		   operators, but not now. */
+		   of text we test depends entirely on the operator. */
 
-		for(n = 0, so = stdops; 
-		    n < STDOP_CNT &&
-		    (FEQ(so->name, so->namel) ||
-		     !(so->ctx_l == context || (so->ctx_l == ANY && 
-		     			        context != NON)));
+		for(so = stdops, n = 0, e = STDOP_CNT; 
+		    n < e && (FEQ(so->name, so->namel) ||
+			      !(so->ctx_l == context || (so->ctx_l == ANY && context != NON)));
 		    n++, so++);
 
-		if (n < STDOP_CNT) {
+		if (n == e) {
+
+			/* The terms and operators we will search for next are
+			 * all tokens that consist entirely of letters, digits,
+			 * dashes and colons. If the last character is a colon,
+			 * we leave that out, so that 'int:=3' works as
+			 * expected. */
+
+			/* Tokenize now if we haven't already done so and
+			 * complain if that doesn't give us anything. */
+
+			if (len == -1) {
+				len = TOKEN();
+				if (len && I(len - 1) == ':') len--; 
+			}
+			if (len <= 0) goto operror;
+
+			/* Search the named operator table for first
+			 * operator matching name and context. */
+
+			for(so = namedops, n = 0, e = NAMEDOP_CNT; 
+			    n < e && (EQ(len, so->name, so->namel) ||
+				      !(so->ctx_l == context || (so->ctx_l == ANY && context != NON)));
+			    n++, so++);
+
+		}
+
+		/* Short and named operators are processed the same way */
+
+		if (n < e) {
 
 			/* Check precedence. Associating right means don't
 			   break if same precedence but continue recursing. */
@@ -502,10 +564,9 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 			/* Recurse, specifying required right hand type */
 			if (so->cvt_r != NON) {
 				l = reccompile(m, ifaces, source, sourcel,
-					       &o, bufsize - ret, 
-					       so->prec, so->cvt_r, ctx_item, 0,
-					       fullpath, pathl, 
-					       funchead, functail);
+					&o, bufsize - ret, 
+					so->prec, so->cvt_r, ctx_item, 0, 0,
+					fullpath, pathl, funchead, functail);
 				if (l == -1) return -1;	
 				ret += l;
 			}
@@ -515,33 +576,72 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 				o->op = so->op; o++; ret += sizeof(INSN);
 			}
 
-			/* Set new context; reset ctx_item if none (comma) */
+			/* Set new context */
 			context = so->newctx;
-			if (context == NON) ctx_item = 0;
 			continue;
 		}
 
-		/* The searches below only apply in context NON. So if we are
-		 * still looking for a typed binary or postfix operator here,
-		 * it won't be found, and we can complain that we don't 
-		 * know acalssbout the operator used in the source.
-		 *
-		 * The terms and unary prefix operators we look for are all
-		 * named and consist entirely of letters, digits, dashes and
-		 * colons and are ended by any other character. If the last
-		 * character of the name is a colon or a dash, we leave that
-		 * out, so that 'int:=3' and 'int-3' work as expected.  This
-		 * should be changed, because there are attributes that have
-		 * digits after a dash, and even values that *start* with
-		 * digits and then a dash, eg. 1200-75-BPS. Yuck. We'll fix
-		 * that some other time, by always testing for constants first
-		 * if there's an attribute in context and we expect a term. */
 
-		if (context != NON || !((len == -1 && (len = SPN())), len)) {
-			msg(F_LANG, L_ERR, "compile: ERROR: no valid operator found after term! (missing comma?)\n");
-			goto error; 
+		/*
+		 * NAMED SPECIAL OPERATORS
+		 */
+
+		/* These boolean operators use short-circuit evaluation, so
+		 * 'A and B' evaluates B only if A is true, and 
+		 * 'A or B', evaluates B only if A is false. 
+		 *
+		 * They always return the last evaluated term as the result, so
+		 * 'A and B' returns either false or B, and 
+		 * 'A or B' returns A if that's true and B otherwise. 
+		 *
+		 * Note that these have a very low precedence, and 'or' is
+		 * just below 'and'. */
+
+		if (context != NON && !EQ(len, "and", 3)) {
+			if (precedence >= 5) break;
+			S(2 + (c == 'a'));
+			saved_o = o;
+			o->op = OP_JMPZ; o++;
+			o->op = OP_POP; o++; 
+			/* Funky feature: 'and' does not require the left hand
+			 * term to be any specific type, because to the VM, a
+			 * false int looks exactly like a false str, so if we
+			 * return that, then that's what the stack has anyway.
+			 * In all cases, the returned context is the right 
+			 * hand's type. */
+			l = reccompile(m, ifaces, source, sourcel,
+				       &o, bufsize - ret, 
+				       5, ANY, ctx_item, &context, 0,
+				       fullpath, pathl, funchead, functail);
+			if (l == -1) return -1;	
+			saved_o->imm.d.disp = l / sizeof(INSN) + 1;
+			ret += l + 2 * sizeof(INSN);
+			continue;
 		}
-		if (I(len - 1) == ':' || I(len - 1) == '-') l--;
+
+		if (context != NON && !EQ(len, "or", 2)) {
+			if (precedence >= 4) break;
+			S(2);
+			saved_o = o;
+			o->op = OP_JMPNZ; o++;
+			o->op = OP_POP; o++; 
+			/* Require the right term to be the same type as the
+			 * left: pass current context as required. */
+			l = reccompile(m, ifaces, source, sourcel,
+				       &o, bufsize - ret, 
+				       4, context, ctx_item, 0, 0,
+				       fullpath, pathl, funchead, functail);
+			if (l == -1) return -1;	
+			saved_o->imm.d.disp = l / sizeof(INSN) + 1;
+			ret += l + 2 * sizeof(INSN);
+			continue;
+		}
+
+		/* All next searches only apply in context NON; if we are still
+		 * looking for a typed binary or postfix operator here, it
+		 * won't be found, and we can already complain. */
+
+		if (context != NON) goto operror;
 
 		/* Include files */
 
@@ -551,13 +651,13 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 			str = str_getword(&src, pstr(" \n\r\t,"));
 			*source = src.p; *sourcel = src.l;
 
-			/* We don't autoconvert the subexpression in the file
-			 * but inherit the context here */
+			/* We are at NON, start subexpression at NON (we always
+			 * do), expect NON, and stay at NON. Yeah. */
+
 			l = reccompilefile(m, ifaces, 
 					   fullpath, pathl, str.p, str.l, 
 					   &o, bufsize - ret, 
-					   0, ANY, 0, &context,
-					   funchead, functail);
+					   0, NON, 0, 0, funchead, functail);
 			if (l == -1) return -1;	
 			ret += l;
 			continue;
@@ -583,12 +683,13 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 			func_add(funchead, functail, str.p, str.l, o);
 
 			/* Compile at precedence 28 (unary prefix ops),
-			 * make function auto-convert stack top to int
-			 * in case there's no explicit return (requiring int),
-			 * reset ctx_item, don't care about returned ctx. */
+			 * make function body auto-convert to int, in
+			 * in case there's no explicit return (which also
+			 * converts to int */
+	
 			l = reccompile(m, ifaces, source, sourcel,
 				       &o, bufsize - ret, 
-				       28, INT, 0, 0,
+				       28, INT, 0, 0, 0,
 				       fullpath, pathl, funchead, functail);
 			if (l == -1) return -1;	
 
@@ -608,18 +709,19 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 		/*
 		 * FUNCTION CALLS
 		 *
-		 * These behave as unary prefix operators that
-		 * evaluate but pop (NON) the right hand side and return 
-		 * the integer returned by the function itself.
+		 * These behave as unary prefix operators that evaluate and 
+		 * discard an optional right hand term, and return the integer
+		 * returned by the function itself. We set allow_non to allow
+		 * reccompile to terminate early, without obtaining a term.
 		 */
 
 		saved_o = func_getaddr(funchead, *source, len);
 		if (saved_o) {
-			if (precedence >= 28) break;
+			if (precedence > 28) break;
 			S(len);
 			l = reccompile(m, ifaces, source, sourcel, 
 				       &o, bufsize - ret, 
-				       28, NON, 0, 0,
+				       28, NON, 0, 0, 1,
 				       fullpath, pathl, 
 				       funchead, functail);
 			if (l == -1) return -1;	
@@ -638,19 +740,21 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 		 * INTERFACE CALLS
 		 */
 
-		/* Search callable interfaces, which behave as unary
-		 * prefix operators that evaluate but pop the right hand
-		 * side and keep the context at 'NON' (term required) */
+		/* Search callable interfaces. Interface calls evaluate
+		 * the right hand term, if any, then call the interface,
+		 * and return the right hand term, or NON if none was
+		 * given. We set allow_non to allow reccompile to terminate
+		 * without obtaining a term at all. */
 
 		for(ifc = ifaces; 
 		    ifc && EQ(len, ifc->name, ifc->namel);
 		    ifc = ifc->next);
 		if (ifc) {
-			if (precedence >= 28) break;
+			if (precedence > 28) break;
 			S(len);
 			l = reccompile(m, ifaces, source, sourcel, 
 				       &o, bufsize - ret, 
-				       28, NON, 0, 0,
+				       28, ANY, 0, &context, 1,
 				       fullpath, pathl, 
 				       funchead, functail);
 			if (l == -1) return -1;	
@@ -663,6 +767,7 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 			continue;
 		}
 
+
 		/*
 		 * DICTIONARY TERMS
 		 */
@@ -671,36 +776,34 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 						      &o->imm.i.ref);
 		if (o->imm.i.item) {
 			S(len);
-
-			/* This is a hack. If the reference doesn't specify
-			 * a list, record the case of the first letter of the
-			 * reference so that the default can be dependent on
-			 * that. 
-			 *
-			 * A lowercase letter causes the opcode's default
-			 * target list to be inverted. Used for 'str', etc.,
-			 * that you generally set on the request list and test
-			 * on the reply list. 
-			 *
-			 * This will go away. It successfully decreases the
-			 * number of prefixes in behaviour files, but is a bit
-			 * too hard to automate mentally. */
-
-			if ((o->imm.i.ref.have & AVR_HAVE_LIST) == 0 && 
-			    ISLWR(o->imm.i.item->name[0]))
-				o->imm.i.ref.list = AVR_LIST_REP;
-				/* note that we don't set AVR_HAVE_LIST */
-
 			context = o->imm.i.item->val_type;
 			ctx_item = o->imm.i.item;
 			o->op = OP_PUSHAV; o++; ret += sizeof(INSN);
 			continue;
 		}
 
-		/* Error. See if we have an itemized context or not. */
+
+		/*
+		 * NOTHING FOUND
+		 */
+
+		/* Special funky feature: if the parent allows us to
+		 * push nothing, then just return without an error; let the
+		 * parent deal with whatever we got and whatever's left ahead.
+		 * Note that we're always in NON if we even get here. 
+		 *
+		 * Using this, we allow binary operators to terminate optional
+		 * subexpressions. Use sparingly. We only use optional
+		 * subexpressions for interface- and function calls. */
+
+		if (allow_non) break;
+
+		/* Error. Message depends on whether we have an attribute 
+		 * context in scope. */
+
 		if (ctx_item) {
 			msg(F_LANG, L_ERR, "compile: ERROR: Unknown "
-					   "term or operator '%s' (for "
+					   "term or unary operator '%s' (for "
 					   "%s:%s:%s)!\n",
 			    DBGCVT(len), ctx_item->spc->name,
 			    meta_getvndbynr(m, ctx_item->vnd)->name,
@@ -708,30 +811,37 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 		}
 		else {
 			msg(F_LANG, L_ERR, "compile: ERROR: Unknown "
-					   "term or operator '%s'!\n", 
+					   "term or unary operator '%s'!\n", 
 			    DBGCVT(len));
 		}
 		goto error;
 	}
 
-	/* Add conversion operator if requested context doesn't match current.
-	 * If we were requested a term (req_ctx != NON) and we or our children 
-	 * didn't push any, complain. If we were requested no term, and we
-	 * or our children pushed one, pop it. Otherwise, autoconvert. */
+	/* If parent didn't want a term (req_ctx NON) and we pushed one, pop
+	 * it.
+	 *
+	 * If parent required a term (req_ctx not NON and allow_non clear)
+	 * flag an error if we didn't push any. 
+	 *
+	 * If parent required a term and we pushed one, autoconvert if
+	 * necessary.
+	 */
 
-	if (req_ctx != context) {
-		if (context == NON) {
-			msg(F_LANG, L_ERR, "compile: ERROR: term expected!\n");
-			goto error;
-		}
-		if (req_ctx == NON) {
+	if (req_ctx == NON) {
+		if (context != NON) {
 			o->op = OP_POP; 
 			o++, ret += sizeof(INSN);
 		}
-		else {
+	}
+	else {					/* req_ctx != NON */
+		if (context != NON) {
 			o->op = cvtop(context, req_ctx);
 			if (o->op != OP_NOP) o++, ret += sizeof(INSN);
 		}
+		else if (allow_non == 0) {	/* context NON */
+			msg(F_LANG, L_ERR, "compile: ERROR: term expected!\n");
+			goto error;
+		}
 	}
 
 	*buf = o;
@@ -739,6 +849,9 @@ static ssize_t reccompile(META *m, IFACE *ifaces,
 	D1(msg(F_LANG, L_DEBUG, "reccompile: Returned.\n"));
 	return ret;
 
+operror:
+	msg(F_LANG, L_ERR, "compile: ERROR: no valid operator found!"
+			   " (missing comma?)\n");
 error:
 	l = *sourcel; s = " <eof>";
 	if (l > 20) l = 20, s = "...";
@@ -763,8 +876,8 @@ static ssize_t reccompilefile(META *m, IFACE *ifaces,
 			      FUNC_T **funchead, FUNC_T **functail)
 {
 	ssize_t ret;
-	char *fullpath;
-	STR_T text, src;
+	char *fullpath, *freep;
+	STR_T text;
 
 	ret = -1;
 
@@ -787,17 +900,17 @@ static ssize_t reccompilefile(META *m, IFACE *ifaces,
 		return -1;
 	}
 
-	src = text;
-	ret = reccompile(m, ifaces, &src.p, &src.l, buf, bufsize, 
-			 precedence, req_ctx, ctx_item, ret_ctx,
+	freep = text.p;
+	ret = reccompile(m, ifaces, &text.p, &text.l, buf, bufsize, 
+			 precedence, req_ctx, ctx_item, ret_ctx, 0,
 			 fullpath, pathl, funchead, functail);
-	free(text.p);
+	free(freep);
 	free(fullpath);
+
 	return ret;
 }
 
 
-
 /* Called by other modules */
 
 ssize_t lang_compilefile(META *m, IFACE *ifaces, 
@@ -814,6 +927,8 @@ ssize_t lang_compilefile(META *m, IFACE *ifaces,
 			     0, NON, 0, 0,
 			     funchead, functail);
 
+	/* Terminate with a HALT to be sure */
+
 	if (ret != -1 && ret < bufsize - sizeof(INSN)) {
 		buf->op = OP_HALT; ret += sizeof(INSN);
 	}
@@ -831,7 +946,10 @@ ssize_t lang_compile(META *m, IFACE *ifaces,
 {
 	ssize_t ret;
 	ret = reccompile(m, ifaces, &source, &sourcel, &buf, bufsize,
-			 0, NON, 0, 0, 0, 0, funchead, functail);
+			 0, NON, 0, 0, 0,
+			 0, 0, funchead, functail);
+
+	/* Terminate with a HALT to be sure */
 
 	if (ret != -1 && ret < bufsize - sizeof(INSN)) {
 		buf->op = OP_HALT; ret += sizeof(INSN);
